# Standard Library
import logging
from datetime import UTC, datetime, timedelta
from itertools import islice

# Third Party
from sqlalchemy import extract, func, select, update
from sqlalchemy.engine import Row
from sqlalchemy.orm import Query, Session

# First Party
from resc_backend.constants import (
    AUDIT_AUTOMATED_AUDITOR,
    AUDIT_AUTOMATED_COMMENT,
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    MAX_RECORDS_PER_PAGE_LIMIT,
)
from resc_backend.db.model import DBaudit, DBfinding, DBrepository, DBVcsInstance
from resc_backend.resc_web_service.schema.audit import AuditFinding
from resc_backend.resc_web_service.schema.auditor_metric import AuditorMetric
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.time_period import TimePeriod

logger = logging.getLogger(__name__)

YEAR = "year"
MONTH = "month"
WEEK = "week"
DAY = "day"


def create_audits(
    db_connection: Session,
    finding_ids: set[int],
    auditor: str,
    status: FindingStatus,
    comment: str = "",
) -> list[DBaudit]:
    """
        Audit finding, updating the status and comment
    :param db_connection:
        Session of the database connection
    :param finding_ids:
        List of id of the finding to audit
    :param auditor:
        identifier of the person performing the audit action
    :param status:
        audit status to set, type FindingStatus
    :param comment:
        audit comment to set
    :return: DBaudit
        The output will contain the audit that was created
    """
    # Iterate over those ids by chunk.
    # This is necessary because SQL tends to crash when you do IN with more than 1000 values.
    # source: trust me bro.
    iterator = iter(finding_ids)
    db_audits = []
    while chunk := list(islice(iterator, 1000)):
        db_connection.execute(update(DBaudit).where(DBaudit.finding_id.in_(chunk)).values(is_latest=False))
        db_audits_created = []

        # Loop around the findings and audit one by one.
        for finding_id in chunk:
            db_audit = DBaudit(
                finding_id=finding_id,
                auditor=auditor,
                status=status,
                comment=comment,
                timestamp=datetime.now(UTC),
                is_latest=True,
            )
            db_audits_created.append(db_audit)

        # Insert new Audits by chunk.
        db_connection.add_all(db_audits_created)

        db_audits.extend(db_audits_created)

    # Commit the change.
    db_connection.commit()

    return db_audits


def create_automated_audits(db_connection: Session, findings_ids: list[int], status: FindingStatus) -> list[DBaudit]:
    """
        Create automated audit for a list of findings.

    Args:
        db_connection (Session): Session of the database connection
        findings_ids (list[int]): list of id to audit
        status (FindingStatus): status to apply

    Returns:
        list[DBaudit]: newly created audits
    """

    # Iterate over those ids by chunk.
    # This is necessary because SQL tends to crash when you do IN with more than 1000 values.
    # source: trust me bro.
    iterator = iter(findings_ids)
    db_audits = []
    while chunk := list(islice(iterator, 1000)):
        db_connection.execute(update(DBaudit).where(DBaudit.finding_id.in_(chunk)).values(is_latest=False))
        db_audits_created = [DBaudit.create_automated(finding_id, status) for finding_id in chunk]
        db_connection.add_all(db_audits_created)
        db_audits.extend(db_audits_created)

    db_connection.commit()

    logger.debug(f"Automated audit of {len(db_audits)} findings.")

    return db_audits


def clear_outdated_no_longer_outdated(db_connection: Session, findings_ids: list[int]) -> None:
    """
        Remove outdated status from findings which have been automatically added

    Args:
        db_connection (Session): Session of the database connection
        findings_ids (list[int]): list of id to audit
    """
    iterator = iter(findings_ids)
    while chunk := list(islice(iterator, 1000)):
        query = db_connection.query(DBaudit)
        query = query.where(DBaudit.finding_id.in_(chunk))
        query = query.where(DBaudit.auditor == AUDIT_AUTOMATED_AUDITOR)
        query = query.where(DBaudit.comment == AUDIT_AUTOMATED_COMMENT)
        query.delete(synchronize_session=False)

    fix_last_audit(db_connection, findings_ids)


def get_finding_audits(
    db_connection: Session,
    finding_id: int,
    skip: int = 0,
    limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT,
) -> list[DBaudit]:
    """
        Get Audit entries for finding
    :param db_connection:
        Session of the database connection
    :param finding_id:
        id of the finding to audit
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :return: [DBaudit]
        The output will contain the list of audit items for the given finding
    """
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit
    query = select(DBaudit).where(DBaudit.finding_id == finding_id)
    query = query.order_by(DBaudit.id_.desc()).offset(skip).limit(limit_val)
    finding_audits = db_connection.execute(query).scalars().all()
    return finding_audits


def get_finding_audits_count(db_connection: Session, finding_id: int) -> int:
    """
        Get count of Audit entries for finding
    :param db_connection:
        Session of the database connection
    :param finding_id:
        id of the finding to audit
    :return: total_count
        count of audit entries
    """
    query = select(func.count(DBaudit.id_)).where(DBaudit.finding_id == finding_id)
    total_count = db_connection.execute(query).scalars().one()
    return total_count


def get_audit_count_by_auditor_over_time(db_connection: Session, weeks: int = 13) -> list[Row]:
    """
        Retrieve count audits by auditor over time for given weeks
    :param db_connection:
        Session of the database connection
    :param weeks:
        optional, filter on last n weeks, default 13
    :return: count_over_time
        list of rows containing audit count over time per week
    """
    last_nth_week_date_time = datetime.now(UTC) - timedelta(weeks=weeks)

    query = (
        db_connection.query(
            extract(YEAR, DBaudit.timestamp).label(YEAR),
            extract(WEEK, DBaudit.timestamp).label(WEEK),
            DBaudit.auditor,
            func.count(DBaudit.id_).label("audit_count"),
        )
        .filter(
            (extract(YEAR, DBaudit.timestamp) > extract(YEAR, last_nth_week_date_time))
            | (
                (extract(YEAR, DBaudit.timestamp) == extract(YEAR, last_nth_week_date_time))
                & (extract(WEEK, DBaudit.timestamp) >= extract(WEEK, last_nth_week_date_time))
            )
        )
        .group_by(
            extract(YEAR, DBaudit.timestamp).label(YEAR),
            extract(WEEK, DBaudit.timestamp).label(WEEK),
            DBaudit.auditor,
        )
        .order_by(
            extract(YEAR, DBaudit.timestamp).label(YEAR),
            extract(WEEK, DBaudit.timestamp).label(WEEK),
            DBaudit.auditor,
        )
    )
    finding_audits = query.all()

    return finding_audits


def get_personal_audit_count(db_connection: Session, auditor: str, time_period: TimePeriod) -> int:
    """
        Get count of Audit entries for finding
    :param db_connection:
        Session of the database connection
    :param auditor:
        id of the auditor
    :param time_period:
        period for which to retrieve the audit counts
    :return: total_count
        count of audit entries
    """
    date_today = datetime.now(UTC)

    total_count = db_connection.query(func.count(DBaudit.id_))

    if time_period in (time_period.DAY, time_period.MONTH, time_period.YEAR):
        total_count = total_count.filter(extract(YEAR, DBaudit.timestamp) == extract(YEAR, date_today))

        if time_period in (time_period.DAY, time_period.MONTH):
            total_count = total_count.filter(extract(MONTH, DBaudit.timestamp) == extract(MONTH, date_today))

            if time_period == time_period.DAY:
                total_count = total_count.filter(extract(DAY, DBaudit.timestamp) == extract(DAY, date_today))

    if time_period in (time_period.WEEK, time_period.LAST_WEEK):
        date_last_week = datetime.now(UTC) - timedelta(weeks=1)
        date_week = date_last_week if time_period == time_period.LAST_WEEK else date_today
        total_count = total_count.filter(extract(YEAR, DBaudit.timestamp) == extract(YEAR, date_week))
        total_count = total_count.filter(extract(WEEK, DBaudit.timestamp) == extract(WEEK, date_week))

    total_count = total_count.filter(DBaudit.auditor == auditor).scalar()
    return total_count


def get_audit_stats_count(db_connection: Session, auditor: str | None) -> list[AuditorMetric]:
    """Retrieve the stats True Positive, False Positive etc... per auditor

    Args:
        db_connection (Session): Session of the database connection
        auditor (str): id of the auditor

    Returns:
        list: List of AuditorMetrics
    """
    true_positive: Query = select(DBaudit.auditor, func.count(DBaudit.auditor).label("true_positive"))
    true_positive = true_positive.where(DBaudit.status == FindingStatus.TRUE_POSITIVE)
    true_positive = true_positive.group_by(DBaudit.auditor)
    true_positive = true_positive.subquery()

    false_positive: Query = select(DBaudit.auditor, func.count(DBaudit.auditor).label("false_positive"))
    false_positive = false_positive.where(DBaudit.status == FindingStatus.FALSE_POSITIVE)
    false_positive = false_positive.group_by(DBaudit.auditor)
    false_positive = false_positive.subquery()

    clarification_required: Query = select(DBaudit.auditor, func.count(DBaudit.auditor).label("clarification_required"))
    clarification_required = clarification_required.where(DBaudit.status == FindingStatus.CLARIFICATION_REQUIRED)
    clarification_required = clarification_required.group_by(DBaudit.auditor)
    clarification_required = clarification_required.subquery()

    not_accessible: Query = select(DBaudit.auditor, func.count(DBaudit.auditor).label("not_accessible"))
    not_accessible = not_accessible.where(DBaudit.status == FindingStatus.NOT_ACCESSIBLE)
    not_accessible = not_accessible.group_by(DBaudit.auditor)
    not_accessible = not_accessible.subquery()

    outdated: Query = select(DBaudit.auditor, func.count(DBaudit.auditor).label("outdated"))
    outdated = outdated.where(DBaudit.status == FindingStatus.OUTDATED)
    outdated = outdated.group_by(DBaudit.auditor)
    outdated = outdated.subquery()

    not_analyzed: Query = select(DBaudit.auditor, func.count(DBaudit.auditor).label("not_analyzed"))
    not_analyzed = not_analyzed.where(DBaudit.status == FindingStatus.NOT_ANALYZED)
    not_analyzed = not_analyzed.group_by(DBaudit.auditor)
    not_analyzed = not_analyzed.subquery()

    total: Query = select(DBaudit.auditor, func.count(DBaudit.auditor).label("total"))
    total = total.group_by(DBaudit.auditor)
    total = total.subquery()

    query = select(
        DBaudit.auditor,
        func.coalesce(true_positive.c.true_positive, 0).label("true_positive"),
        func.coalesce(false_positive.c.false_positive, 0).label("false_positive"),
        func.coalesce(clarification_required.c.clarification_required, 0).label("clarification_required"),
        func.coalesce(not_accessible.c.not_accessible, 0).label("not_accessible"),
        func.coalesce(outdated.c.outdated, 0).label("outdated"),
        func.coalesce(not_analyzed.c.not_analyzed, 0).label("not_analyzed"),
        func.coalesce(total.c.total, 0).label("total"),
    )
    query = query.join(true_positive, true_positive.c.auditor == DBaudit.auditor, isouter=True)
    query = query.join(false_positive, false_positive.c.auditor == DBaudit.auditor, isouter=True)
    query = query.join(clarification_required, clarification_required.c.auditor == DBaudit.auditor, isouter=True)
    query = query.join(not_accessible, not_accessible.c.auditor == DBaudit.auditor, isouter=True)
    query = query.join(outdated, outdated.c.auditor == DBaudit.auditor, isouter=True)
    query = query.join(not_analyzed, not_analyzed.c.auditor == DBaudit.auditor, isouter=True)
    query = query.join(total, total.c.auditor == DBaudit.auditor, isouter=True)

    if auditor is not None:
        query = query.where(DBaudit.auditor == auditor)

    query = query.distinct()

    return db_connection.execute(query).all()


def fix_last_audit(db_connection: Session, finding_ids: list[int]) -> None:
    # Iterate over those ids by chunk.
    # This is necessary because SQL tends to crash when you do IN with more than 1000 values.
    # source: trust me bro.
    finding_ids_iterator = iter(finding_ids)
    while chunk := list(islice(finding_ids_iterator, 1000)):
        # Create a sub query with group by on finding.
        max_audit_subquery: Query = select(DBaudit.finding_id, func.max(DBaudit.id_).label("audit_id"))
        max_audit_subquery = max_audit_subquery.where(DBaudit.finding_id.in_(chunk))
        max_audit_subquery = max_audit_subquery.group_by(DBaudit.finding_id)
        max_audit_subquery = max_audit_subquery.subquery()

        # Select the id from previously selected tupples.
        latest_audits_query = select(DBaudit.id_)
        latest_audits_query = latest_audits_query.join(max_audit_subquery, max_audit_subquery.c.audit_id == DBaudit.id_)
        latest_audits = db_connection.execute(latest_audits_query).scalars().all()
        query = update(DBaudit).where(DBaudit.id_.in_(latest_audits)).values(is_latest=True)
        db_connection.execute(query)

    db_connection.commit()


def revert_last_audit(db_connection: Session, finding_ids: list[int], status: FindingStatus | None) -> None:
    """
    Revert the last audit for a set of findings with specific status

    Args:
        db_connection (Session): Database session
        finding_ids (list[int]): List of findings
        status (FindingStatus | None): status to remove.
    """
    iterator = iter(finding_ids)
    while chunk := list(islice(iterator, 1000)):
        query = db_connection.query(DBaudit)
        query = query.where(DBaudit.finding_id.in_(chunk))
        query = query.where(DBaudit.is_latest == True)  # noqa: E712
        if status is not None:
            query = query.where(DBaudit.status == status)
        query.delete(synchronize_session=False)

    fix_last_audit(db_connection, finding_ids)


def _audit_list_filtering(
    query: Query,
    auditor: str | None,
    from_date: datetime | None,
    to_date: datetime | None,
    status: list[FindingStatus] | None,
    is_latest: bool | None,
) -> Query:
    """
    Limit the query with the following conditions:

    Args:
        db_connection (Session): Database session
        auditor (str | None) : optional restriction on the auditor
        from_date (datetime | None): optional restriciton on the dates
        to_date (datetime | None): optional restriciton on the dates
        status (list[FindingStatus] | None): optional restrictions on the statuses
        is_latest (bool | None): only consider latest.
    """

    query = query.where(DBaudit.auditor != AUDIT_AUTOMATED_AUDITOR)
    if auditor:
        query = query.where(DBaudit.auditor == auditor)
    if from_date:
        query = query.where(DBaudit.timestamp > from_date)
    if to_date:
        query = query.where(DBaudit.timestamp < to_date)
    if status and len(status) > 0:
        query = query.where(DBaudit.status.in_(status))
    if is_latest:
        query = query.where(DBaudit.is_latest == True)  # noqa: E712
    return query


def get_audits(
    db_connection: Session,
    skip: int,
    limit: int,
    auditor: str | None,
    from_date: datetime | None,
    to_date: datetime | None,
    status: list[FindingStatus] | None,
    is_latest: bool | None,
):
    """
    Fetch the recent audits given some conditions

    Args:
        db_connection (Session): Database session
        skip (int): skip in the data query
        limit (int): limit in the data query
        auditor (str | None) : optional restriction on the auditor
        from_date (datetime | None): optional restriciton on the dates
        to_date (datetime | None): optional restriciton on the dates
        status (list[FindingStatus] | None): optional restrictions on the statuses
        is_latest (bool | None): only consider latest.
    """

    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit

    query = db_connection.query(
        DBaudit.id_.label("audit_id"),
        DBaudit.status,
        DBaudit.auditor,
        DBaudit.comment,
        DBaudit.timestamp,
        DBaudit.is_latest,
        DBfinding.id_.label("finding_id"),
        DBfinding.file_path,
        DBfinding.line_number,
        DBfinding.column_start,
        DBfinding.column_end,
        DBfinding.commit_id,
        DBfinding.commit_message,
        DBfinding.commit_timestamp,
        DBfinding.author,
        DBfinding.email,
        DBfinding.is_dir_scan,
        DBfinding.rule_name,
        DBfinding.event_sent_on,
        DBVcsInstance.provider_type.label("vcs_provider"),
        DBrepository.project_key,
        DBrepository.repository_name,
        DBrepository.repository_url,
    )
    query = query.join(DBfinding, DBfinding.id_ == DBaudit.finding_id)
    query = query.join(DBrepository, DBfinding.repository_id == DBrepository.id_)
    query = query.join(DBVcsInstance, DBrepository.vcs_instance == DBVcsInstance.id_)
    query = _audit_list_filtering(query, auditor, from_date, to_date, status, is_latest)

    query = query.order_by(DBaudit.timestamp.desc())
    query = query.offset(skip).limit(limit_val)
    findings: list[AuditFinding] = query.all()

    return findings


def get_total_audits_count(
    db_connection: Session,
    auditor: str | None,
    from_date: datetime | None,
    to_date: datetime | None,
    status: list[FindingStatus] | None,
    is_latest: bool | None,
):
    """
    Get the totall count of the recent audits given some conditions

    Args:
        db_connection (Session): Database session
        auditor (str | None) : optional restriction on the auditor
        from_date (datetime | None): optional restriciton on the dates
        to_date (datetime | None): optional restriciton on the dates
        status (list[FindingStatus] | None): optional restrictions on the statuses
        is_latest (bool | None): only consider latest.
    """
    query = select(func.count(DBaudit.id_))
    query = query.join(DBfinding, DBfinding.id_ == DBaudit.finding_id)
    query = _audit_list_filtering(query, auditor, from_date, to_date, status, is_latest)

    total_count = db_connection.execute(query).scalars().one()
    return total_count



# Standard Library
from datetime import datetime

# Third Party
from fastapi import APIRouter, Depends, Query, status
from fastapi_cache.decorator import cache
from sqlalchemy.orm import Session

# First Party
from resc_backend.constants import (
    CACHE_NAMESPACE_FINDING,
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    ERROR_MESSAGE_500,
    ERROR_MESSAGE_503,
    FINDINGS_TAG,
    REDIS_CACHE_EXPIRE,
    RWS_ROUTE_AUDITS,
)
from resc_backend.resc_web_service.crud import audit as audit_crud
from resc_backend.resc_web_service.dependencies import get_db_connection
from resc_backend.resc_web_service.schema.audit import AuditFinding
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.pagination_model import PaginationModel

router = APIRouter(prefix=f"{RWS_ROUTE_AUDITS}", tags=[FINDINGS_TAG])


@router.get(
    "",
    response_model=PaginationModel[AuditFinding],
    summary="Get audits",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the past audits"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_FINDING, expire=REDIS_CACHE_EXPIRE)
def get_all_audits(
    skip: int = Query(default=0, ge=0),
    limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
    auditor: str | None = Query(None),
    from_date: datetime | None = Query(None),
    to_date: datetime | None = Query(None),
    status: list[FindingStatus] | None = Query(None),
    is_latest: bool | None = Query(None),
    db_connection: Session = Depends(get_db_connection),
) -> PaginationModel[AuditFinding]:
    """
        Retrieve all audits objects paginated

    - **db_connection**: Session of the database connection
    - **skip**: Integer amount of records to skip to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **auditor**: String to filter which auditor to audit.
    - **from_date**: DateTime to filter from which we look at (oldest)
    - **to_date**: DateTime to filter to which we look at (youngest)
    - **status**: Finding status to filter on
    - **is_latest**: Whether to only consider latest audits.
    - **return**: [AuditFinding]
        The output will contain a PaginationModel containing the list of AuditFinding type objects,
        or an empty list if no audits was found
    """
    audits = audit_crud.get_audits(
        db_connection,
        skip=skip,
        limit=limit,
        auditor=auditor,
        from_date=from_date,
        to_date=to_date,
        status=status,
        is_latest=is_latest,
    )

    total_audits = audit_crud.get_total_audits_count(
        db_connection, auditor=auditor, from_date=from_date, to_date=to_date, status=status, is_latest=is_latest
    )

    return PaginationModel[AuditFinding](data=audits, total=total_audits, limit=limit, skip=skip)



# Standard Library
import datetime
from typing import Annotated

# Third Party
from pydantic import BaseModel, ConfigDict, Field, HttpUrl, StringConstraints, model_validator

# First Party
from resc_backend.constants import MAX_RECORDS_PER_PAGE_LIMIT
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders


class AuditMultiple(BaseModel):
    finding_ids: Annotated[
        list[Annotated[int, Field(gt=0)]], Field(min_length=1, max_length=MAX_RECORDS_PER_PAGE_LIMIT)
    ]
    status: FindingStatus
    comment: Annotated[str, StringConstraints(max_length=255)]


class AuditRead(BaseModel):
    id_: Annotated[int, Field(gt=0)]
    status: FindingStatus
    auditor: Annotated[str, StringConstraints(max_length=250)]
    comment: Annotated[str, StringConstraints(max_length=255)] | None = None
    timestamp: datetime.datetime
    model_config = ConfigDict(from_attributes=True)


class AuditFinding(BaseModel):
    audit_id: Annotated[int, Field(gt=0)]
    status: FindingStatus
    auditor: Annotated[str, StringConstraints(max_length=250)]
    comment: Annotated[str, StringConstraints(max_length=255)] | None = None
    timestamp: datetime.datetime
    is_latest: bool

    finding_id: Annotated[int, Field(gt=0)]
    file_path: str
    line_number: Annotated[int, Field(gt=-1)]
    column_start: Annotated[int, Field(gt=-1)]
    column_end: Annotated[int, Field(gt=-1)]
    commit_id: Annotated[str, StringConstraints(max_length=120)]
    commit_message: str
    commit_timestamp: datetime.datetime
    author: Annotated[str, StringConstraints(max_length=200)]
    email: Annotated[str, StringConstraints(max_length=100)]
    rule_name: Annotated[str, StringConstraints(max_length=200)]
    # rule_pack: Annotated[str, StringConstraints(max_length=100)]
    project_key: Annotated[str, StringConstraints(min_length=1, max_length=100)]
    repository_name: Annotated[str, StringConstraints(min_length=1, max_length=100)]
    repository_url: HttpUrl
    vcs_provider: VCSProviders
    is_dir_scan: bool

    commit_url: Annotated[str, StringConstraints(min_length=1)] | None = None

    @staticmethod
    def build_bitbucket_commit_url(
        repository_url: str,
        repository_name: str,
        project_key: str,
        file_path: str,
        commit_id: str,
        line_number: int,
        is_dir_scan: bool,
    ) -> str:
        arr = repository_url.split("/")
        if len(arr) >= 3:
            repo_base_url = arr[0] + "//" + arr[2]
        else:
            repo_base_url = repository_url

        if is_dir_scan:
            return (
                f"{repo_base_url}/projects/{project_key}/repos/"
                f"{repository_name}/browse/{file_path}?at={commit_id}#{line_number}"
            )

        return (
            f"{repo_base_url}/projects/{project_key}/repos/"
            f"{repository_name}/commits/{commit_id}#{file_path}?t={line_number}"
        )

    @staticmethod
    def build_ado_commit_url(
        repository_url: str,
        file_path: str,
        commit_id: str,
        line_number: int,
        is_dir_scan: bool,
    ) -> str:
        if is_dir_scan:
            return (
                f"{repository_url}?version=GC{commit_id}&path=/{file_path}&line={line_number}&lineEnd={line_number + 1}"
                "&lineStartColumn=1&lineEndColumn=1&type=2&lineStyle=plain"
            )

        return (
            f"{repository_url}/commit/{commit_id}?path=/{file_path}&line={line_number}&lineEnd={line_number + 1}"
            "&lineStartColumn=1&lineEndColumn=1&type=2&lineStyle=plain"
        )

    @staticmethod
    def build_github_commit_url(repository_url: str, file_path: str, commit_id: str) -> str:
        github_commit_url = f"{repository_url}/commit/{commit_id}?path=/{file_path}"
        return github_commit_url

    @model_validator(mode="before")
    def build_commit_url(cls, values: (tuple | dict)) -> dict:  # noqa: N805
        # We are dealing with a namedtuple (i.e. from a database query)
        if hasattr(values, "_asdict"):
            values = values._asdict()

        if "status" not in values or values["status"] is None:
            values["status"] = FindingStatus.NOT_ANALYZED.value
        if "comment" not in values or values["comment"] is None:
            values["comment"] = ""
        if values["vcs_provider"] == VCSProviders.BITBUCKET:
            values["commit_url"] = cls.build_bitbucket_commit_url(
                repository_url=values["repository_url"],
                repository_name=values["repository_name"],
                project_key=values["project_key"],
                file_path=values["file_path"],
                commit_id=values["commit_id"],
                line_number=values["line_number"],
                is_dir_scan=values["is_dir_scan"],
            )
        elif values["vcs_provider"] == VCSProviders.AZURE_DEVOPS:
            values["commit_url"] = cls.build_ado_commit_url(
                repository_url=values["repository_url"],
                file_path=values["file_path"],
                commit_id=values["commit_id"],
                line_number=values["line_number"],
                is_dir_scan=values["is_dir_scan"],
            )

        elif values["vcs_provider"] == VCSProviders.GITHUB_PUBLIC:
            values["commit_url"] = cls.build_github_commit_url(
                repository_url=values["repository_url"],
                file_path=values["file_path"],
                commit_id=values["commit_id"],
            )
        else:
            raise NotImplementedError(f"Unsupported VCSProvider: {values['vcs_provider']}")

        return values

    model_config = ConfigDict(from_attributes=True)



# Standard Library

# Third Party
from typing import Annotated

from pydantic import BaseModel, ConfigDict, Field, StringConstraints


class RuleAllowListBase(BaseModel):
    description: Annotated[str, StringConstraints(max_length=2000)] | None = None
    regexes: str | None = None
    paths: str | None = None
    commits: str | None = None
    stop_words: str | None = None


class RuleAllowListCreate(RuleAllowListBase):
    @classmethod
    def create_from_base_class(cls, base_object: RuleAllowListBase):
        return cls(**(dict(base_object)))


class RuleAllowList(RuleAllowListBase):
    pass


class RuleAllowListRead(RuleAllowListBase):
    id_: Annotated[int, Field(gt=0)]
    model_config = ConfigDict(from_attributes=True)



# Third Party

from typing import Annotated

from pydantic import BaseModel, ConfigDict, Field, StringConstraints

from resc_backend.constants import MAX_RECORDS_PER_PAGE_LIMIT

# First Party
from resc_backend.resc_web_service.schema.status_count import StatusCount


class RuleFindingCountModel(BaseModel):
    """
    :param Generic[Model]:
        Type of the object in the data list
    """

    rule_name: Annotated[str, StringConstraints(max_length=100)]
    finding_count: Annotated[int, Field(gt=-1)] = 0
    finding_statuses_count: Annotated[
        list[StatusCount], Field(min_length=None, max_length=MAX_RECORDS_PER_PAGE_LIMIT)
    ] = []
    model_config = ConfigDict(from_attributes=True)



# Standard Library
import logging
import re

# Third Party
import tomlkit
from fastapi import APIRouter, Depends, File, HTTPException, Query, UploadFile, status
from fastapi.responses import FileResponse
from fastapi_cache.decorator import cache
from packaging.version import Version
from sqlalchemy.orm import Session

# First Party
from resc_backend.constants import (
    CACHE_NAMESPACE_FINDING,
    CACHE_NAMESPACE_RULE,
    CACHE_NAMESPACE_RULE_PACK,
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    ERROR_MESSAGE_500,
    ERROR_MESSAGE_503,
    REDIS_CACHE_EXPIRE,
    RULE_PACKS_TAG,
    RWS_ROUTE_MARK_AS_OUTDATED,
    RWS_ROUTE_RULE_PACKS,
    RWS_ROUTE_RULES,
    RWS_ROUTE_TAGS,
    RWS_ROUTE_VERSIONS,
)
from resc_backend.resc_web_service.cache_manager import CacheManager
from resc_backend.resc_web_service.crud import audit as audit_crud
from resc_backend.resc_web_service.crud import finding as finding_crud
from resc_backend.resc_web_service.crud import rule as rule_crud
from resc_backend.resc_web_service.crud import rule_pack as rule_pack_crud
from resc_backend.resc_web_service.crud import rule_tag as rule_tag_crud
from resc_backend.resc_web_service.dependencies import get_db_connection
from resc_backend.resc_web_service.helpers.resc_swagger_models import (
    Model400,
    Model403,
    Model404,
    Model409,
    Model422,
)
from resc_backend.resc_web_service.helpers.rule import (
    create_toml_dictionary,
    get_mapped_global_allow_list_obj,
    map_dictionary_to_rule_allow_list_object,
    validate_uploaded_file_and_read_content,
)
from resc_backend.resc_web_service.helpers.toml import create_toml_rule_file
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.pagination_model import PaginationModel
from resc_backend.resc_web_service.schema.rule import RuleCreate, RuleRead
from resc_backend.resc_web_service.schema.rule_allow_list import RuleAllowList
from resc_backend.resc_web_service.schema.rule_pack import RulePackCreate, RulePackRead, RulePackVersion

router = APIRouter(prefix=f"{RWS_ROUTE_RULE_PACKS}", tags=[RULE_PACKS_TAG])

logger = logging.getLogger(__name__)


@router.get(
    f"{RWS_ROUTE_VERSIONS}",
    response_model=PaginationModel[RulePackRead],
    summary="Get rule packs",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the rule-packs"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_RULE_PACK, expire=REDIS_CACHE_EXPIRE)
def get_rule_packs(
    version: str | None = Query(None, pattern=r"^\d+(?:\.\d+){2}$"),
    active: bool | None = Query(None, description="Filter on active rule packs"),
    skip: int = Query(default=0, ge=0),
    limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
    db_connection: Session = Depends(get_db_connection),
) -> PaginationModel[RulePackRead]:
    """
        Retrieve rule packs

    - **db_connection**: Session of the database connection
    - **version**: Optional, filter on rule pack version
    - **active**: Optional, filter on active rule pack
    - **skip**: Integer amount of records to skip, to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **return**: [RulePackRead]
        The output will contain a PaginationModel containing the list of RulePackRead type objects,
        or an empty list if no rule pack was found
    """
    rule_packs = rule_pack_crud.get_rule_packs(
        db_connection=db_connection,
        version=version,
        active=active,
        skip=skip,
        limit=limit,
    )
    total_rule_packs_count = rule_pack_crud.get_total_rule_packs_count(
        db_connection=db_connection, version=version, active=active
    )
    return PaginationModel[RulePackRead](data=rule_packs, total=total_rule_packs_count, limit=limit, skip=skip)


@router.get(
    "",
    summary="Download rule pack in TOML format",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Download the rule-pack in TOML format"},
        404: {
            "model": Model404,
            "description": "No rule-pack of version <version_id> found",
        },
        422: {
            "model": Model422,
            "description": "Version <version_id> is not a valid semantic version",
        },
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def download_rule_pack_toml_file(
    rule_pack_version: str | None = Query(None, pattern=r"^\d+(?:\.\d+){2}$"),
    db_connection: Session = Depends(get_db_connection),
) -> FileResponse:
    """
        Download rule pack in TOML format

    - **db_connection**: Session of the database connection
    - **version**: Optional, filter on rule pack version
    - **return**: [FileResponse] The output returns rule pack file downloaded in TOML format
    """
    if not rule_pack_version:
        logger.info("rule pack version not specified, downloading the currently active version")
    rule_pack_from_db = read_rule_pack(version=rule_pack_version, db_connection=db_connection)
    if rule_pack_from_db:
        version = rule_pack_from_db.version
        rules = rule_crud.get_rules_by_rule_pack_version(db_connection=db_connection, rule_pack_version=version)
        rule_tag_names = rule_tag_crud.get_rule_tag_names_by_rule_pack_version(
            db_connection=db_connection, rule_pack_version=version
        )
        global_allow_list = rule_crud.get_global_allow_list_by_rule_pack_version(
            db_connection=db_connection, rule_pack_version=version
        )
        generated_toml_dict = create_toml_dictionary(version, rules, global_allow_list, rule_tag_names)
    else:
        raise HTTPException(status_code=404, detail=f"No rule pack found with version {rule_pack_version}")

    toml_file = create_toml_rule_file(generated_toml_dict)
    return FileResponse(toml_file.name, filename="RESC-SECRETS-RULE.toml")


@router.post(
    "",
    summary="Upload rule pack in TOML format",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Upload the rule-pack in TOML format"},
        400: {
            "model": Model400,
            "description": "No properties defined for rule allow list",
        },
        409: {
            "model": Model409,
            "description": "Rule-pack version <version_id> already exists",
        },
        422: {
            "model": Model422,
            "description": "Version <version_id> is not a valid semantic version",
        },
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def upload_rule_pack_toml_file(
    version: str = Query(pattern=r"^\d+(?:\.\d+){2}$"),
    rule_file: UploadFile = File(...),
    db_connection: Session = Depends(get_db_connection),
) -> dict:
    """
        Upload TOML rule pack to database

    - **db_connection**: Session of the database connection
    - **version**: Version of the rule pack to be uploaded
    - **rule_file**: TOML rule pack file to be uploaded
    - **return**: dict The output returns uploaded rule pack name in dictionary format
    """
    content = validate_uploaded_file_and_read_content(rule_file)
    toml_rule_dictionary = tomlkit.loads(content)

    # Check if rule pack version exists
    rule_pack_from_db = read_rule_pack(version=version, db_connection=db_connection)
    if rule_pack_from_db:
        raise HTTPException(
            status_code=409,
            detail=f"Unable to process rules. Rule pack version {version} already exists",
        )

    # Insert in to RULE_ALLOW_LIST for storing global allow list
    created_global_rule_allow_list = None
    global_allow_list: RuleAllowList = get_mapped_global_allow_list_obj(toml_rule_dictionary)
    if global_allow_list:
        created_global_rule_allow_list = create_rule_allow_list(
            rule_allow_list=global_allow_list, db_connection=db_connection
        )
        if not created_global_rule_allow_list.id_:
            logger.warning("Creating global rule allow list failed with an error")

    # # Insert in to RULE_PACK
    global_allow_list_id = (
        created_global_rule_allow_list.id_
        if created_global_rule_allow_list and created_global_rule_allow_list.id_
        else None
    )

    # Determine if uploaded rule pack needs to be activated
    current_newest_rule_pack = rule_pack_crud.get_newest_rule_pack(db_connection)
    activate_uploaded_rule_pack = determine_uploaded_rule_pack_activation(version, current_newest_rule_pack)

    rule_pack = RulePackCreate(
        version=version,
        active=activate_uploaded_rule_pack,
        global_allow_list=global_allow_list_id,
    )
    created_rule_pack_version = create_rule_pack_version(rule_pack=rule_pack, db_connection=db_connection)
    if created_rule_pack_version.version and activate_uploaded_rule_pack:
        # Update older rule packs to inactive
        rule_pack_crud.make_older_rule_packs_to_inactive(latest_rule_pack_version=version, db_connection=db_connection)
    else:
        logger.warning("Creating rule pack failed with an error")

    # Insert in to RULES
    if created_rule_pack_version and created_rule_pack_version.version:
        insert_rules(
            version=version,
            toml_rule_dictionary=toml_rule_dictionary,
            db_connection=db_connection,
        )

        # Clear cache related to rule-pack
        await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_RULE)
        await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_RULE_PACK)
        await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_FINDING)
    return {"filename": rule_file.filename}


def read_rule_pack(version: str | None = None, db_connection: Session = Depends(get_db_connection)) -> RulePackRead:
    """
        Read active rule pack from database
    :param version:
        optional, version of the rule pack to be fetched else latest rule pack version will be fetched
    :param db_connection:
        Session of the database connection
    :return: RulePackRead
        The output returns RulePackRead type object
    """
    if version:
        regex = re.compile(r"^\d+(?:\.\d+){2}$")
        if not re.fullmatch(regex, version):
            raise HTTPException(
                status_code=422,
                detail=f"Version {version} is not a valid semantic version",
            )
    db_rule_pack = rule_pack_crud.get_rule_pack(db_connection=db_connection, version=version)
    return db_rule_pack


def create_rule_pack_version(rule_pack: RulePackCreate, db_connection: Session = Depends(get_db_connection)):
    """
        Create rule pack version in database
    :param rule_pack:
        RulePackCreate object to be created
    :param db_connection:
        Session of the database connection
    """
    return rule_pack_crud.create_rule_pack_version(db_connection=db_connection, rule_pack=rule_pack)


def create_rule_allow_list(rule_allow_list: RuleAllowList, db_connection: Session = Depends(get_db_connection)):
    """
        Create rule allow list in database
    :param rule_allow_list:
        RuleAllowList object to be created
    :param db_connection:
        Session of the database connection
    """
    if (
        rule_allow_list.paths
        or rule_allow_list.commits
        or rule_allow_list.stop_words
        or rule_allow_list.description
        or rule_allow_list.regexes
    ):
        return rule_crud.create_rule_allow_list(db_connection=db_connection, rule_allow_list=rule_allow_list)
    raise HTTPException(status_code=400, detail="No properties defined for rule allow list")


def insert_rules(
    version: str,
    toml_rule_dictionary: dict,
    db_connection: Session = Depends(get_db_connection),
):
    """
        Create rules in database
    :param version:
        version of the rule pack
    :param toml_rule_dictionary:
        rule pack toml in dictionary format
    :param db_connection:
        Session of the database connection
    """
    if "rules" in toml_rule_dictionary:
        rule_list = toml_rule_dictionary.get("rules")
        for rule in rule_list:
            keywords = None
            rule_name = rule["id"] if "id" in rule else None
            description = rule["description"] if "description" in rule else None
            regex = rule["regex"] if "regex" in rule else None
            entropy = rule["entropy"] if "entropy" in rule else None
            secret_group = rule["secretGroup"] if "secretGroup" in rule else None
            path = rule["path"] if "path" in rule else None
            comment = rule["comment"] if "comment" in rule else None

            if "keywords" in rule:
                keyword_array = rule["keywords"]
                keywords = ",".join(keyword_array)

            # Insert in to RULE_ALLOW_LIST for storing individual rule specific  allow list
            allow_list = rule["allowlist"] if "allowlist" in rule else None
            rule_allow_list_obj = map_dictionary_to_rule_allow_list_object(allow_list)
            created_rule_allow_list = None
            if rule_allow_list_obj:
                created_rule_allow_list = create_rule_allow_list(
                    rule_allow_list=rule_allow_list_obj, db_connection=db_connection
                )

            # Insert in to RULES
            if allow_list and created_rule_allow_list and created_rule_allow_list.id_:
                created_allow_list_id = created_rule_allow_list.id_
            else:
                created_allow_list_id = None
            rule_obj = RuleCreate(
                rule_pack=version,
                allow_list=created_allow_list_id,
                rule_name=rule_name,
                description=description,
                entropy=entropy,
                secret_group=secret_group,
                regex=regex,
                path=path,
                keywords=keywords,
                comment=comment,
            )
            created_rule = rule_crud.create_rule(rule=rule_obj, db_connection=db_connection)
            if not created_rule.id_:
                logger.warning(f"Creating rule failed for Rule: {rule_name}")
            if "tags" in rule:
                _ = rule_tag_crud.create_rule_tag(
                    db_connection=db_connection,
                    rule_id=created_rule.id_,
                    tags=rule["tags"],
                )


@router.get(
    f"{RWS_ROUTE_TAGS}",
    response_model=list[str],
    summary="Get rule packs' tags",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the tags related to a rule-pack[s]"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_RULE_PACK, expire=REDIS_CACHE_EXPIRE)
def get_rule_packs_tags(
    versions: list[str] | None = Query(None, alias="version", title="version"),
    db_connection: Session = Depends(get_db_connection),
) -> list[str]:
    """
        Retrieve rule pack related tags

    :param db_connection: Session of the database connection
    :param versions: Optional, filter on rule pack version, if not provided filter on active.
    :return: List[str]
        The output will contain a list of tags related to one or more rule-packs.
    """
    if not versions:
        active_rule_pack = rule_pack_crud.get_current_active_rule_pack(db_connection=db_connection)
        if not active_rule_pack:
            raise HTTPException(status_code=500, detail="No currently active rule pack.")
        versions = [active_rule_pack.version]
    rule_packs_tags = rule_pack_crud.get_rule_packs_tags(db_connection=db_connection, versions=versions)
    return rule_packs_tags


def determine_uploaded_rule_pack_activation(
    requested_rule_pack_version: str, latest_rule_pack_from_db: RulePackRead
) -> bool:
    """
        Determine if rule pack needs to be activated
    :param requested_rule_pack_version:
        version of the rule pack uploaded
    :param latest_rule_pack_from_db:
        latest rule pack in RulePackRead object
    :return: boolean
        The output returns true if rule pack needs to be activated else returns false
    """
    if latest_rule_pack_from_db:
        if Version(latest_rule_pack_from_db.version) < Version(requested_rule_pack_version):
            logger.info(
                f"Uploaded rule pack is of version '{requested_rule_pack_version}', using it to replace "
                f"'{latest_rule_pack_from_db.version}' as the active one."
            )
            activate_uploaded_rule_pack = True
        else:
            if not latest_rule_pack_from_db.active:
                logger.info(
                    f"There is already a more recent rule pack present in the database "
                    f"'{latest_rule_pack_from_db.version}', but it is set to inactive. "
                    f"Activating the uploaded rule pack '{requested_rule_pack_version}'"
                )
                activate_uploaded_rule_pack = True
            else:
                logger.info(
                    f"Uploaded rule pack is of version '{requested_rule_pack_version}', the existing rule pack "
                    f"'{latest_rule_pack_from_db.version}' is kept as the active one."
                )
                activate_uploaded_rule_pack = False
    else:
        logger.info(
            f"No existing rule pack found, So activating the uploaded rule pack '{requested_rule_pack_version}'"
        )
        activate_uploaded_rule_pack = True
    return activate_uploaded_rule_pack


@router.post(
    f"{RWS_ROUTE_MARK_AS_OUTDATED}",
    summary="Mark rule pack as outdated",
    response_model=dict,
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Rulepack successfully updated"},
        403: {
            "model": Model403,
            "description": "Rule-pack version <version_id> is active",
        },
        404: {
            "model": Model404,
            "description": "Rule-pack version <version_id> does not exists",
        },
        422: {
            "model": Model422,
            "description": "Version <version_id> is not a valid semantic version",
        },
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def mark_rule_pack_as_outdated(
    rulepack: RulePackVersion,
    db_connection: Session = Depends(get_db_connection),
) -> dict:
    """
        Mark rule pack as outdated.

        We fetch all the findings which are not in any scans that are in younger rule pack.
        We mark those findings as out dated.
        e.g.

        Finding 1 - Scan 1 - Rulepack 1
        Finding 1 - Scan 2 - Rulepack 2
        Finding 1 - Scan 3 - Rulepack 3
        Finding 2 - Scan 1 - Rulepack 1
        Finding 2 - Scan 2 - Rulepack 2
        Finding 3 - Scan 3 - Rulepack 3

        Marking Rule pack 2 as outdated will mark finding 2 as outdated: 1 and 3 are still found in rule pack 3
        Marking Rule pack 1 as outdated will mark no findings as outdated: 1, 2 and 3 are found in rule packs 2 and 3.

    - **db_connection**: Session of the database connection
    - **version**: Version of the rule pack to be marked
    - **return**: int The number of findings marked.
    """
    rule_packs = rule_pack_crud.get_rule_packs(
        db_connection=db_connection,
        version=rulepack.version,
    )
    if len(rule_packs) != 1:
        raise HTTPException(status_code=404, detail="Rule pack not found.")

    if rule_packs[0].active:
        raise HTTPException(status_code=403, detail="Rule pack is active.")

    findings: list[int] = finding_crud.get_untriaged_finding_for_old_rulepacks(
        db_connection=db_connection, version=rulepack.version
    )
    audits = audit_crud.create_automated_audits(
        db_connection=db_connection, findings_ids=findings, status=FindingStatus.OUTDATED
    )

    return {"audited": len(audits)}


@router.get(
    f"/{{rule_pack_version}}{RWS_ROUTE_RULES}",
    response_model=RuleRead,
    summary="Get unique rule from rule pack",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve the rule data for a rule pack"},
        404: {"model": Model404, "description": "Rule <rule_name> for <rule_pack_version> not found"},
        422: {"model": Model422, "description": "rule_pack_version and rule_name required"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_RULE, expire=REDIS_CACHE_EXPIRE)
def get_rule_from_rule_pack(
    rule_pack_version: str,
    rule_name: str = Query(alias="rule_name", title="RuleName"),
    db_connection: Session = Depends(get_db_connection),
) -> RuleRead:
    """
        Retrieve the rule data from a rule_name and rule_pack

    - **db_connection**: Session of the database connection
    - **rule_pack_version**: filter on rule pack version
    - **rule_name**: filter on rule pack version
    - **return**: List[str] The output will contain a list of strings of unique rules in the findings table
    """
    db_rule = rule_crud.get_rule_by_rule_name_and_rule_pack_version(
        db_connection=db_connection, rule_name=rule_name, rule_pack_version=rule_pack_version
    )
    if db_rule is None:
        raise HTTPException(status_code=404, detail="Rule not found")

    return RuleRead.model_validate(db_rule)



import datetime
from typing import Annotated

# Third Party
from pydantic import BaseModel, ConfigDict, Field, HttpUrl, StringConstraints


class SimpleRepository(BaseModel):
    id: Annotated[str, StringConstraints(min_length=1, max_length=100)]
    name: Annotated[str, StringConstraints(min_length=1, max_length=100)]


class ActiveRepositories(BaseModel):
    project_key: Annotated[str, StringConstraints(min_length=1, max_length=100)]
    repositories: Annotated[list[SimpleRepository], Field(min_length=0)]
    vcs_instance_name: Annotated[str, StringConstraints(max_length=200)]


class RepositoryBase(BaseModel):
    project_key: Annotated[str, StringConstraints(min_length=1, max_length=100)]
    repository_id: Annotated[str, StringConstraints(min_length=1, max_length=100)]
    repository_name: Annotated[str, StringConstraints(min_length=1, max_length=100)]
    repository_url: HttpUrl
    deleted_at: datetime.datetime | None = None
    vcs_instance: Annotated[int, Field(gt=0)]


class Repository(RepositoryBase):
    pass


class RepositoryCreate(RepositoryBase):
    pass


class RepositoryRead(RepositoryBase):
    id_: Annotated[int, Field(gt=0)]
    model_config = ConfigDict(from_attributes=True)
